# Copyright (c) 2017 NEC Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from networking_odl.common import constants as odl_const
from networking_odl.db import models
from networking_odl.journal import journal
from networking_odl.tests.unit import base_v2


class OpenDaylightJournalThreadTest(base_v2.OpenDaylightTestCase):
    def setUp(self):
        super(OpenDaylightJournalThreadTest, self).setUp()
        self.journal = journal.OpenDaylightJournalThread()
        self.addCleanup(self.cleanup)

    @staticmethod
    def cleanup():
        journal.MAKE_URL.clear()

    def test_json_data(self):
        object_type = 'testobject'
        data = 'testdata'
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        self.assertEqual("%ss" % object_type, self.journal._json_data(row)[1])

    def test_json_data_customized_url(self):
        object_type = 'randomtestobject'
        data = 'testdata'
        journal.register_url_builder(object_type, lambda row: row.object_type)
        row = models.OpenDaylightJournal(object_type=object_type,
                                         object_uuid=uuidutils.generate_uuid(),
                                         operation=odl_const.ODL_CREATE,
                                         data=data)

        url_param = self.journal._json_data(row)
        self.assertEqual(object_type, url_param[1])



# Copyright 2016 Intel Corporation.
# Copyright 2016 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.tests.unit import test_base_db


class DbTestCase(test_base_db.ODLBaseDbTestCase):

    UPDATE_ROW = [odl_const.ODL_NETWORK, 'id', odl_const.ODL_UPDATE,
                  {'test': 'data'}]
    model = models.OpenDaylightJournal

    def setUp(self):
        super(DbTestCase, self).setUp()

    def _create_row(self):
        db.create_pending_row(self.db_session, *self.UPDATE_ROW)
        rows = db.get_all_db_rows(self.db_session)
        self.assertEqual(1, len(rows))
        return rows[0]

    def test_equal_created_at(self):
        row = self._create_row()
        got = self.db_session.query(self.model).filter_by(
            created_at=row.created_at).all()
        self.assertEqual(1, len(got))

    def test_get_by_primary_key_filter(self):
        row = self._create_row()
        # NOTE(manjeets) as seqnum is primary key so there would be
        # exactly one row created.
        query = self.db_session.query(models.OpenDaylightJournal)
        got = query.filter_by(seqnum=row.seqnum).one()
        self.assertEqual(row, got)

    def test_compare_created_at(self):
        row = self._create_row()
        created_at = row.created_at + datetime.timedelta(minutes=1)
        got = self.db_session.query(self.model).filter(
            self.model.created_at < created_at).all()
        self.assertEqual(1, len(got))



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base

from networking_odl.common import utils


class TestUtils(base.DietTestCase):

    # TODO(manjeets) remove this test once neutronify is
    # consolidated with make_plural
    def test_neutronify(self):
        self.assertEqual('a-b-c', utils.neutronify('a_b_c'))

    def test_neutronify_empty(self):
        self.assertEqual('', utils.neutronify(''))

    def test_make_url_object_in_resource_map(self):
        url_object = utils.make_url_object('policy')
        self.assertEqual('qos/policies', url_object)

    def test_make_url_object_conversion(self):
        self.assertEqual('networks', utils.make_url_object('network'))
        self.assertEqual('l2-gateways', utils.make_url_object('l2_gateway'))



#
# Copyright (C) 2017 NEC, Corp.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lbaas.drivers import driver_base
from neutron_lib.plugins import constants as nlib_const

from networking_odl.common import constants as odl_const
from networking_odl.journal import full_sync
from networking_odl.journal import journal

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = logging.getLogger(__name__)

LBAAS_RESOURCES = {
    odl_const.ODL_LOADBALANCER: odl_const.ODL_LOADBALANCERS,
    odl_const.ODL_LISTENER: odl_const.ODL_LISTENERS,
    odl_const.ODL_POOL: odl_const.ODL_POOLS,
    odl_const.ODL_MEMBER: odl_const.ODL_MEMBERS,
    odl_const.ODL_HEALTHMONITOR: odl_const.ODL_HEALTHMONITORS
}


class OpenDaylightManager(driver_base.LoadBalancerBaseDriver):
    """OpenDaylight LBaaS Driver for the V2 API

    This code is the backend implementation for the OpenDaylight
    LBaaS V2 driver for OpenStack Neutron.
    """

    @log_helpers.log_method_call
    def __init__(self, driver, obj_type):
        LOG.debug("Initializing OpenDaylight LBaaS driver")
        super(OpenDaylightManager, self).__init__(driver)
        self.journal = journal.OpenDaylightJournalThread()
        self.obj_type = obj_type
        full_sync.register(nlib_const.LOADBALANCERV2, LBAAS_RESOURCES)

    def _journal_record(self, context, obj_type, obj_id, operation, obj):
        obj_type = ("lbaas/%s" % obj_type)
        journal.record(context, obj_type, obj_id, operation, obj)
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    @driver_base.driver_op
    def create(self, context, obj):
        self._journal_record(context, self.obj_type, obj.id,
                             odl_const.ODL_CREATE, obj)

    @log_helpers.log_method_call
    @driver_base.driver_op
    def update(self, context, obj):
        self._journal_record(context, self.obj_type, obj.id,
                             odl_const.ODL_UPDATE, obj)

    @log_helpers.log_method_call
    @driver_base.driver_op
    def delete(self, context, obj):
        self._journal_record(context, self.obj_type, obj.id,
                             odl_const.ODL_DELETE, obj)


class ODLLoadBalancerManager(OpenDaylightManager,
                             driver_base.BaseLoadBalancerManager):

    @log_helpers.log_method_call
    def __init__(self, driver):
        super(ODLLoadBalancerManager, self).__init__(
            driver, odl_const.ODL_LOADBALANCER)

    @log_helpers.log_method_call
    @driver_base.driver_op
    def refresh(self, context, lb):
        # TODO(lijingjing): implement this method
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        pass

    @log_helpers.log_method_call
    @driver_base.driver_op
    def stats(self, context, lb):
        # TODO(rajivk): implement this method
        pass


class ODLListenerManager(OpenDaylightManager,
                         driver_base.BaseListenerManager):

    @log_helpers.log_method_call
    def __init__(self, driver):
        super(ODLListenerManager, self).__init__(
            driver, odl_const.ODL_LISTENER)


class ODLPoolManager(OpenDaylightManager,
                     driver_base.BasePoolManager):

    @log_helpers.log_method_call
    def __init__(self, driver):
        super(ODLPoolManager, self).__init__(
            driver, odl_const.ODL_POOL)


class ODLMemberManager(OpenDaylightManager,
                       driver_base.BaseMemberManager):

    @log_helpers.log_method_call
    def __init__(self, driver):
        super(ODLMemberManager, self).__init__(
            driver, odl_const.ODL_MEMBER)

        journal.register_url_builder(odl_const.ODL_MEMBER,
                                     self.lbaas_member_url_builder)

    @staticmethod
    def lbaas_member_url_builder(row):
        return ("lbaas/pools/%s/member" % row.data.pool.id)


class ODLHealthMonitorManager(OpenDaylightManager,
                              driver_base.BaseHealthMonitorManager):

    @log_helpers.log_method_call
    def __init__(self, driver):
        super(ODLHealthMonitorManager, self).__init__(
            driver, odl_const.ODL_HEALTHMONITOR)



# Copyright (c) 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.plugins.ml2 import config
from neutron.plugins.ml2.drivers import type_vxlan  # noqa
from neutron.tests.unit.plugins.ml2 import test_plugin

from networking_odl.common import config as odl_config


class TestODLFullStackBase(test_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = ['logger', 'opendaylight']
    _extension_drivers = ['port_security']

    def setUp(self):
        config.cfg.CONF.set_override('extension_drivers',
                                     self._extension_drivers,
                                     group='ml2')
        config.cfg.CONF.set_override('tenant_network_types',
                                     ['vxlan'],
                                     group='ml2')
        config.cfg.CONF.set_override('vni_ranges',
                                     ['1:1000'],
                                     group='ml2_type_vxlan')

        odl_url = 'http://127.0.0.1:8087/controller/nb/v2/neutron'
        odl_config.cfg.CONF.set_override('url',
                                         odl_url,
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('username',
                                         'admin',
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('password',
                                         'admin',
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('port_binding_controller',
                                         'legacy-port-binding',
                                         group='ml2_odl')

        super(TestODLFullStackBase, self).setUp()



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import threading
import time

import mock

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.db import models
from networking_odl.journal import periodic_task
from networking_odl.tests.unit import test_base_db


class PeriodicTaskThreadTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        super(PeriodicTaskThreadTestCase, self).setUp()
        row = models.OpenDaylightPeriodicTask(task='test-maintenance',
                                              state=odl_const.PENDING)
        self.db_session.add(row)
        self.db_session.flush()

        self.thread = periodic_task.PeriodicTask('test-maintenance')
        self.thread.interval = 0.01
        self.addCleanup(self.thread.cleanup)

    def test__execute_op_no_exception(self):
        with mock.patch.object(periodic_task, 'LOG') as mock_log:
            operation = mock.MagicMock()
            operation.__name__ = "test"
            self.thread.register_operation(operation)
            self.thread._execute_op(operation, self.db_context)
            operation.assert_called()
            mock_log.info.assert_called()
            mock_log.exception.assert_not_called()

    def test__execute_op_with_exception(self):
        with mock.patch.object(periodic_task, 'LOG') as mock_log:
            operation = mock.MagicMock(side_effect=Exception())
            operation.__name__ = "test"
            self.thread._execute_op(operation, self.db_context)
            mock_log.exception.assert_called()

    def test_thread_works(self):
        callback_event = threading.Event()
        count = [0]

        def callback_op(**kwargs):
            count[0] += 1

            # The following should be true on the second call, so we're making
            # sure that the thread runs more than once.
            if count[0] > 1:
                callback_event.set()

        self.thread.register_operation(callback_op)
        self.thread.start()

        # Make sure the callback event was called and not timed out
        self.assertTrue(callback_event.wait(timeout=5))

    def test_thread_continues_after_exception(self):
        exception_event = threading.Event()
        callback_event = threading.Event()

        def exception_op(**kwargs):
            if not exception_event.is_set():
                exception_event.set()
                raise Exception()

        def callback_op(**kwargs):
            callback_event.set()

        for op in [exception_op, callback_op]:
            self.thread.register_operation(op)

        self.thread.start()

        # Make sure the callback event was called and not timed out
        self.assertTrue(callback_event.wait(timeout=5))

    def test_multiple_thread_work(self):
        self.thread1 = periodic_task.PeriodicTask('test-maintenance1')
        callback_event = threading.Event()
        callback_event1 = threading.Event()
        self.thread1.interval = 0.01
        self.addCleanup(self.thread1.cleanup)

        def callback_op(**kwargs):
            callback_event.set()

        def callback_op1(**kwargs):
            callback_event1.set()

        self.thread.register_operation(callback_op)
        self.thread.register_operation(callback_op1)
        self.thread.start()
        self.assertTrue(callback_event.wait(timeout=5))

        self.thread1.start()
        self.assertTrue(callback_event1.wait(timeout=5))

    @mock.patch.object(db, "was_periodic_task_executed_recently")
    def test_back_to_back_job(self, mock_status_method):
        callback_event = threading.Event()

        def callback_op(**kwargs):
            callback_event.set()

        self.thread.register_operation(callback_op)
        msg = ("Periodic %s task executed after periodic "
               "interval Skipping execution.")
        with mock.patch.object(periodic_task.LOG, 'info') as mock_log_info:
            mock_status_method.return_value = True
            self.thread.start()
            time.sleep(1)
            mock_log_info.assert_called_with(msg, 'test-maintenance')
            self.assertFalse(callback_event.wait(timeout=1))
            mock_log_info.assert_called_with(msg, 'test-maintenance')
            mock_status_method.return_value = False
            self.assertTrue(callback_event.wait(timeout=2))



#
# Copyright (C) 2016 Ericsson India Global Services Pvt Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import copy

import mock

from networking_odl.l2gateway import driver
from neutron.tests import base


class TestOpenDaylightL2gwDriver(base.DietTestCase):

    def setUp(self):
        self.mocked_odlclient = mock.patch(
            'networking_odl.common.client'
            '.OpenDaylightRestClient.create_client').start().return_value
        self.driver = driver.OpenDaylightL2gwDriver(service_plugin=None,
                                                    validator=None)
        super(TestOpenDaylightL2gwDriver, self).setUp()

    def _get_fake_l2_gateway(self):
        fake_l2_gateway_id = "5227c228-6bba-4bbe-bdb8-6942768ff0f1"
        fake_l2_gateway = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff0f1",
            "name": "test-gateway",
            "devices": [
                {
                    "device_name": "switch1",
                    "interfaces": [
                        {
                            "name": "port1",
                            "segmentation_id": [100]
                        },
                        {
                            "name": "port2",
                            "segmentation_id": [151, 152]
                        }
                    ]
                },
                {
                    "device_name": "switch2",
                    "interfaces": [
                        {
                            "name": "port5",
                            "segmentation_id": [200]
                        },
                        {
                            "name": "port6",
                            "segmentation_id": [251, 252]
                        }
                    ]
                }
            ]
        }
        return fake_l2_gateway_id, fake_l2_gateway

    def _get_fake_l2_gateway_connection(self):
        fake_l2_gateway_connection_id = "5227c228-6bba-4bbe-bdb8-6942768ff02f"
        fake_l2_gateway_connection = {
            "tenant_id": "de0a7495-05c4-4be0-b796-1412835c6820",
            "id": "5227c228-6bba-4bbe-bdb8-6942768ff02f",
            "network_id": "be0a7495-05c4-4be0-b796-1412835c6830",
            "default_segmentation_id": 77,
            "l2_gateway_id": "5227c228-6bba-4bbe-bdb8-6942768ff0f1"
        }
        return fake_l2_gateway_connection_id, fake_l2_gateway_connection

    def test_create_l2_gateway_postcommit(self):
        mocked_sendjson = self.mocked_odlclient.sendjson
        fake_l2gateway_id, fake_l2gateway = self._get_fake_l2_gateway()
        expected = {"l2_gateway": fake_l2gateway}
        self.driver.create_l2_gateway_postcommit(mock.ANY, fake_l2gateway)
        mocked_sendjson.assert_called_once_with('post', driver.L2GATEWAYS,
                                                expected)

    def test_delete_l2_gateway_postcommit(self):
        mocked_trydelete = self.mocked_odlclient.try_delete
        fake_l2gateway_id, fake_l2gateway = self._get_fake_l2_gateway()
        self.driver.delete_l2_gateway_postcommit(mock.ANY, fake_l2gateway_id)
        url = driver.L2GATEWAYS + '/' + fake_l2gateway_id
        mocked_trydelete.assert_called_once_with(url)

    def test_update_l2_gateway_postcommit(self):
        mocked_sendjson = self.mocked_odlclient.sendjson
        fake_l2gateway_id, fake_l2gateway = self._get_fake_l2_gateway()
        expected = {"l2_gateway": fake_l2gateway}
        self.driver.update_l2_gateway_postcommit(mock.ANY, fake_l2gateway)
        url = driver.L2GATEWAYS + '/' + fake_l2gateway_id
        mocked_sendjson.assert_called_once_with('put', url, expected)

    def test_create_l2_gateway_connection_postcommit(self):
        mocked_sendjson = self.mocked_odlclient.sendjson
        (fake_l2gateway_conn_id,
         fake_l2gateway_conn) = self._get_fake_l2_gateway_connection()
        expected_l2gateway_conn = copy.deepcopy(fake_l2gateway_conn)
        expected_l2gateway_conn['gateway_id'] = (
            fake_l2gateway_conn['l2_gateway_id'])
        expected_l2gateway_conn.pop('l2_gateway_id')
        expected = {"l2gateway_connection": expected_l2gateway_conn}
        self.driver.create_l2_gateway_connection_postcommit(
            mock.ANY, fake_l2gateway_conn)
        mocked_sendjson.assert_called_once_with('post',
                                                driver.L2GATEWAY_CONNECTIONS,
                                                expected)

    def test_delete_l2_gateway_connection_postcommit(self):
        mocked_trydelete = self.mocked_odlclient.try_delete
        (fake_l2gateway_conn_id,
         fake_l2gateway_conn) = self._get_fake_l2_gateway_connection()
        url = driver.L2GATEWAY_CONNECTIONS + '/' + fake_l2gateway_conn_id
        self.driver.delete_l2_gateway_connection_postcommit(
            mock.ANY, fake_l2gateway_conn_id)
        mocked_trydelete.assert_called_once_with(url)



# Copyright (c) 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from oslo_utils import uuidutils
from tempest.lib.common.utils import test_utils

from neutron.agent.common import utils
from neutron.tests.common import net_helpers

from networking_odl.tests.fullstack import base


class TestMechDriver(base.TestODLFullStackBase):

    def setUp(self):
        super(TestMechDriver, self).setUp()

    def _get_ovs_system_id(self):
        # Example:
        # ('{system-id="f1487b2f-b103-4ce0-b4ab-1a94258405dd"}\n', '')
        system_id = utils.execute(['ovs-vsctl', 'get', 'Open_Vswitch',
                                   '.', 'external_ids'],
                                  run_as_root=True)
        return re.findall('".*"', system_id)[0]

    def _check_device_existence(self, tap):
        def _callback():
            ports = utils.execute(
                ['ovs-vsctl', 'list-ports', 'br-int'],
                run_as_root=True)

            return bool(re.search(tap, ports))
        return test_utils.call_until_true(_callback, 30, 2)

    def _create_ovs_vif_port(self, bridge, dev, iface_id, mac, instance_id):
        return utils.execute(['ovs-vsctl', 'add-port', bridge, dev,
                              '--', 'set', 'Interface', dev,
                              'external-ids:iface-id=%s' % iface_id,
                              'external-ids:iface-status=active',
                              'external-ids:attached-mac=%s' % mac,
                              'external-ids:vm-uuid=%s' % instance_id,
                              'type=tap'], run_as_root=True)

    def test_port_plugging(self):
        # Step1: create test network
        resp = self._create_network(self.fmt, "test_fullstack_net", True)
        resp = self.deserialize(self.fmt, resp)
        net_id = resp['network']['id']

        # Step2: create port and binding to host
        host_id = self._get_ovs_system_id()
        resp = self._create_port(self.fmt, net_id)
        resp = self.deserialize(self.fmt, resp)
        port_id = resp['port']['id']
        mac = resp['port']['mac_address']
        tap = 'tap' + port_id[:net_helpers.OVSPortFixture.NIC_NAME_LEN - 3]
        data = {'port': {'binding:host_id': host_id}}
        req = self.new_update_request('ports', data, port_id)
        resp = self.deserialize(self.fmt, req.get_response(self.api))
        vif_type = resp['port']['binding:vif_type']
        self.assertEqual('ovs', vif_type)
        self.assertFalse(self._check_device_existence(tap))

        # Step3: plug vif
        self._create_ovs_vif_port('br-int', tap, port_id, mac,
                                  uuidutils.generate_uuid())

        # TODO(manjeets) Add a test case to verify mac
        # in flows
        # Step4: verify device
        self.assertTrue(self._check_device_existence(tap))



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

import mock
import requests

from networking_l2gw.services.l2gateway.common import constants as l2gw_const
from networking_sfc.extensions import flowclassifier as fc_const
from networking_sfc.extensions import sfc as sfc_const
from neutron.services.trunk import constants as t_consts
from neutron_lib.api.definitions import bgpvpn as bgpvpn_const
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory

from networking_odl.bgpvpn import odl_v2 as bgpvpn_driver
from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.l2gateway import driver_v2 as l2gw_driver
from networking_odl.l3 import l3_odl_v2
from networking_odl.lbaas import lbaasv2_driver_v2 as lbaas_driver
from networking_odl.ml2 import mech_driver_v2
from networking_odl.qos import qos_driver_v2 as qos_driver
from networking_odl.sfc.flowclassifier import sfc_flowclassifier_v2
from networking_odl.sfc import sfc_driver_v2 as sfc_driver
from networking_odl.tests import base
from networking_odl.tests.unit import test_base_db
from networking_odl.trunk import trunk_driver_v2 as trunk_driver


class FullSyncTestCase(test_base_db.ODLBaseDbTestCase):
    def setUp(self):
        self.useFixture(
            base.OpenDaylightRestClientGlobalFixture(full_sync._CLIENT))
        super(FullSyncTestCase, self).setUp()

        self._CLIENT = full_sync._CLIENT.get_client()

        for plugin_name in self._get_all_resources():
            mocked = mock.MagicMock()
            if plugin_name == constants.CORE:
                self.plugin = mocked
            elif plugin_name == constants.L3:
                self.l3_plugin = mocked

            directory.add_plugin(plugin_name, mocked)

        self.addCleanup(self.clean_registered_resources)

    @staticmethod
    def _get_all_resources():
        return {
            constants.L3: l3_odl_v2.L3_RESOURCES,
            constants.CORE: mech_driver_v2.L2_RESOURCES,
            constants.LOADBALANCERV2: lbaas_driver.LBAAS_RESOURCES,
            t_consts.TRUNK: trunk_driver.TRUNK_RESOURCES,
            constants.QOS: qos_driver.QOS_RESOURCES,
            sfc_const.SFC_EXT: sfc_driver.SFC_RESOURCES,
            bgpvpn_const.LABEL: bgpvpn_driver.BGPVPN_RESOURCES,
            fc_const.FLOW_CLASSIFIER_EXT:
                sfc_flowclassifier_v2.SFC_FC_RESOURCES,
            l2gw_const.L2GW: l2gw_driver.L2GW_RESOURCES,
        }

    @staticmethod
    def clean_registered_resources():
        full_sync.ALL_RESOURCES = {}

    def test_no_full_sync_when_canary_exists(self):
        full_sync.full_sync(self.db_context)
        self.assertEqual([], db.get_all_db_rows(self.db_session))

    def _mock_l2_resources(self):
        expected_journal = {odl_const.ODL_NETWORK: '1',
                            odl_const.ODL_SUBNET: '2',
                            odl_const.ODL_PORT: '3'}
        network_id = expected_journal[odl_const.ODL_NETWORK]
        self.plugin.get_networks.return_value = [{'id': network_id}]
        self.plugin.get_subnets.return_value = [
            {'id': expected_journal[odl_const.ODL_SUBNET],
             'network_id': network_id}]
        port = {'id': expected_journal[odl_const.ODL_PORT],
                odl_const.ODL_SGS: None,
                'tenant_id': '123',
                'fixed_ips': [],
                'network_id': network_id}
        self.plugin.get_ports.side_effect = ([port], [])
        return expected_journal

    def _filter_out_canary(self, rows):
        return [row for row in rows if row['object_uuid'] !=
                full_sync._CANARY_NETWORK_ID]

    def _test_no_full_sync_when_canary_in_journal(self, state):
        self._mock_canary_missing()
        self._mock_l2_resources()
        db.create_pending_row(self.db_session, odl_const.ODL_NETWORK,
                              full_sync._CANARY_NETWORK_ID,
                              odl_const.ODL_CREATE, {})
        row = db.get_all_db_rows(self.db_session)[0]
        db.update_db_row_state(self.db_session, row, state)

        full_sync.full_sync(self.db_context)

        rows = db.get_all_db_rows(self.db_session)
        self.assertEqual([], self._filter_out_canary(rows))

    def test_no_full_sync_when_canary_pending_creation(self):
        self._test_no_full_sync_when_canary_in_journal(odl_const.PENDING)

    def test_no_full_sync_when_canary_is_processing(self):
        self._test_no_full_sync_when_canary_in_journal(odl_const.PROCESSING)

    @mock.patch.object(db, 'delete_pending_rows')
    @mock.patch.object(full_sync, '_full_sync_needed')
    @mock.patch.object(full_sync, '_sync_resources')
    @mock.patch.object(journal, 'record')
    def test_sync_resource_order(
            self, record_mock, _sync_resources_mock, _full_sync_needed_mock,
            delete_pending_rows_mock):

        full_sync.ALL_RESOURCES = self._get_all_resources()
        _full_sync_needed_mock._full_sync_needed.return_value = True
        context = mock.MagicMock()
        full_sync.full_sync(context)

        _sync_resources_mock.assert_has_calls(
            [mock.call(mock.ANY, mock.ANY, object_type, collection_name)
                for (object_type, collection_name) in [
                    (odl_const.ODL_SG, odl_const.ODL_SGS),
                    (odl_const.ODL_SG_RULE, odl_const.ODL_SG_RULES),
                    (odl_const.ODL_NETWORK, odl_const.ODL_NETWORKS),
                    (odl_const.ODL_SUBNET, odl_const.ODL_SUBNETS),
                    (odl_const.ODL_ROUTER, odl_const.ODL_ROUTERS),
                    (odl_const.ODL_PORT, odl_const.ODL_PORTS),
                    (odl_const.ODL_FLOATINGIP, odl_const.ODL_FLOATINGIPS),
                    (odl_const.ODL_LOADBALANCER, odl_const.ODL_LOADBALANCERS),
                    (odl_const.ODL_LISTENER, odl_const.ODL_LISTENERS),
                    (odl_const.ODL_POOL, odl_const.ODL_POOLS),
                    (odl_const.ODL_MEMBER, odl_const.ODL_MEMBERS),
                    (odl_const.ODL_HEALTHMONITOR,
                        odl_const.ODL_HEALTHMONITORS),
                    (odl_const.ODL_QOS_POLICY, odl_const.ODL_QOS_POLICIES),
                    (odl_const.ODL_TRUNK, odl_const.ODL_TRUNKS),
                    (odl_const.ODL_BGPVPN, odl_const.ODL_BGPVPNS),
                    (odl_const.ODL_BGPVPN_NETWORK_ASSOCIATION,
                        odl_const.ODL_BGPVPN_NETWORK_ASSOCIATIONS),
                    (odl_const.ODL_BGPVPN_ROUTER_ASSOCIATION,
                        odl_const.ODL_BGPVPN_ROUTER_ASSOCIATIONS),
                    (odl_const.ODL_SFC_FLOW_CLASSIFIER,
                        odl_const.ODL_SFC_FLOW_CLASSIFIERS),
                    (odl_const.ODL_SFC_PORT_PAIR,
                        odl_const.ODL_SFC_PORT_PAIRS),
                    (odl_const.ODL_SFC_PORT_PAIR_GROUP,
                        odl_const.ODL_SFC_PORT_PAIR_GROUPS),
                    (odl_const.ODL_SFC_PORT_CHAIN,
                        odl_const.ODL_SFC_PORT_CHAINS),
                    (odl_const.ODL_L2GATEWAY, odl_const.ODL_L2GATEWAYS),
                    (odl_const.ODL_L2GATEWAY_CONNECTION,
                        odl_const.ODL_L2GATEWAY_CONNECTIONS)]])

    def test_client_error_propagates(self):
        class TestException(Exception):
            def __init__(self):
                pass

        self._CLIENT.get.side_effect = TestException()
        self.assertRaises(TestException, full_sync.full_sync, self.db_context)

    def _mock_canary_missing(self):
        get_return = mock.MagicMock()
        get_return.status_code = requests.codes.not_found
        self._CLIENT.get.return_value = get_return

    def _assert_canary_created(self):
        rows = db.get_all_db_rows(self.db_session)
        self.assertTrue(any(r['object_uuid'] == full_sync._CANARY_NETWORK_ID
                            for r in rows))
        return rows

    def _test_full_sync_resources(self, expected_journal):
        self._mock_canary_missing()

        full_sync.full_sync(self.db_context)

        rows = self._assert_canary_created()
        rows = self._filter_out_canary(rows)
        self.assertItemsEqual(expected_journal.keys(),
                              [row['object_type'] for row in rows])
        for row in rows:
            self.assertEqual(expected_journal[row['object_type']],
                             row['object_uuid'])

    def test_full_sync_removes_pending_rows(self):
        db.create_pending_row(self.db_session, odl_const.ODL_NETWORK, "uuid",
                              odl_const.ODL_CREATE, {'foo': 'bar'})
        self._test_full_sync_resources({})

    def test_full_sync_no_resources(self):
        self._test_full_sync_resources({})

    def test_full_sync_l2_resources(self):
        full_sync.ALL_RESOURCES = {constants.CORE: mech_driver_v2.L2_RESOURCES}
        self._test_full_sync_resources(self._mock_l2_resources())

    def _mock_router_port(self, port_id):
        router_port = {'id': port_id,
                       'device_id': '1',
                       'tenant_id': '1',
                       'fixed_ips': [{'subnet_id': '1'}]}
        self.plugin.get_ports.side_effect = ([], [router_port])

    def _mock_l3_resources(self):
        expected_journal = {odl_const.ODL_ROUTER: '1',
                            odl_const.ODL_FLOATINGIP: '2'}
        self.l3_plugin.get_routers.return_value = [
            {'id': expected_journal[odl_const.ODL_ROUTER],
             'gw_port_id': None}]
        self.l3_plugin.get_floatingips.return_value = [
            {'id': expected_journal[odl_const.ODL_FLOATINGIP]}]

        return expected_journal

    def test_full_sync_l3_resources(self):
        full_sync.ALL_RESOURCES = {constants.L3: l3_odl_v2.L3_RESOURCES}
        self._test_full_sync_resources(self._mock_l3_resources())
